#include "paging.h"

page_directory_t* kernel_directory  = 0;

ULONG kmalloc_int(ULONG sz, int align, ULONG *phys)
{
    if (align == 1 && (placement_address & 0xFFFFF000) )
    {
        placement_address &= 0xFFFFF000;
        placement_address += 0x1000;
    }
    if (phys)
    {
        *phys = placement_address;
    }
    ULONG tmp = placement_address;
    placement_address += sz;
    return tmp;
}

ULONG kmalloc_a(ULONG sz)
{
    return kmalloc_int(sz, 1, 0);
}

ULONG kmalloc_ap(ULONG sz, ULONG* phys)
{
    return kmalloc_int(sz, 1, phys);
}

ULONG kmalloc(ULONG sz)
{
    return kmalloc_int(sz, 0, 0);
}

static void set_frame(ULONG frame_addr)
{
    ULONG frame = frame_addr/PAGESIZE;

    ULONG idx = INDEX_FROM_BIT(frame);
    ULONG off = OFFSET_FROM_BIT(frame);
    frames[idx] |= (1 << off);
}

static void clear_frame(ULONG frame_addr)
{
    ULONG frame = frame_addr/PAGESIZE;

    ULONG idx = INDEX_FROM_BIT(frame);
    ULONG off = OFFSET_FROM_BIT(frame);
    frames[idx] &= ~(1 << off);
}

static ULONG first_frame()
{
    ULONG i,j;
    for(i=0; i<INDEX_FROM_BIT(nframes); ++i)
    {
        if(frames[i] != 0xFFFFFFFF)
        {
            for(j=0; j<32; ++j)
            {
                ULONG toTest = 0x1 << j;
                if( !(frames[i] & toTest) )
                    return i*32+j;
            }
        }
    }
} //  warning: control reaches end of non-void function  // problem: return

void alloc_frame(page_t *page, int is_kernel, int is_writeable)
{
    if (page->frame != 0)
    {
        return;
    }
    else
    {
        ULONG idx = first_frame();
        if (idx == (ULONG)-1)
            printformat("alloc_frame: no free frames!!!");
        set_frame(idx*PAGESIZE);
        page->present = 1;
        page->rw    = ( is_writeable == 1 ) ? 1 : 0;
        page->user  = ( is_kernel    == 1 ) ? 0 : 1;
        page->frame = idx;
    }
}

void free_frame(page_t *page)
{
    ULONG frame;
    if (!(frame=page->frame))
    {
        return;
    }
    else
    {
        clear_frame(frame);
        page->frame = 0;
    }
}

void paging_install()
{
    ULONG mem_end_page = PHYSICAL_MEMORY;
    nframes = mem_end_page/PAGESIZE;
    frames  = (ULONG*)kmalloc(INDEX_FROM_BIT(nframes));
    k_memset(frames, 0, INDEX_FROM_BIT(nframes));
    kernel_directory = (page_directory_t*)kmalloc_a(sizeof(page_directory_t));
    k_memset(kernel_directory, 0, sizeof(page_directory_t));
    kernel_directory->physicalAddr = (ULONG)kernel_directory->tablesPhysical;

    int i = 0;
    while (i < placement_address)
    {
        alloc_frame( get_page(i, 1, kernel_directory), 0, 0);
        i += PAGESIZE;
    }

    switch_page_directory(kernel_directory);
}

void switch_page_directory(page_directory_t* dir)
{
    asm volatile("mov %0, %%cr3":: "r"(&dir->tablesPhysical));
    ULONG cr0;
    asm volatile("mov %%cr0, %0": "=r"(cr0));
    cr0 |= 0x80000000; // MSB 32 bit // Enable paging
    asm volatile("mov %0, %%cr0":: "r"(cr0));
}

page_t* get_page(ULONG address, int make, page_directory_t* dir)
{
    address /= PAGESIZE;
    ULONG table_idx = address / 1024;

    if (dir->tables[table_idx]) // If this table is already assigned
    {
        return &dir->tables[table_idx]->pages[address%1024];
    }
    else if(make)
    {
        ULONG tmp;
        dir->tables[table_idx] = (page_table_t*)kmalloc_ap(sizeof(page_table_t), &tmp);
        k_memset(dir->tables[table_idx], 0, PAGESIZE);
        dir->tablesPhysical[table_idx] = tmp | 0x7; // PRESENT, RW, US.
        return &dir->tables[table_idx]->pages[address%1024];
    }
    else
    {
        return 0;
    }
}

static page_table_t *clone_table(page_table_t *src, ULONG *physAddr)
{
    page_table_t *table = (page_table_t*)kmalloc_ap(sizeof(page_table_t), physAddr);
    k_memset(table, 0, sizeof(page_directory_t));

    int i;
    for (i = 0; i < 1024; i++)
    {
        if (!src->pages[i].frame)
            continue;
        alloc_frame(&table->pages[i], 0, 0);
        if (src->pages[i].present) table->pages[i].present = 1;
        if (src->pages[i].rw)      table->pages[i].rw = 1;
        if (src->pages[i].user)    table->pages[i].user = 1;
        if (src->pages[i].accessed)table->pages[i].accessed = 1;
        if (src->pages[i].dirty)   table->pages[i].dirty = 1;

        // Physically copy the data across. This assembler function is in process.asm.
        copy_page_physical(src->pages[i].frame*PAGESIZE, table->pages[i].frame*PAGESIZE);
    }
    return table;
}

page_directory_t *clone_directory(page_directory_t *src)
{
    ULONG phys;
    page_directory_t *dir = (page_directory_t*)kmalloc_ap(sizeof(page_directory_t), &phys);
    k_memset(dir, 0, sizeof(page_directory_t));
    ULONG offset = (ULONG)dir->tablesPhysical - (ULONG)dir;
    dir->physicalAddr = phys + offset;

    int i;
    for (i = 0; i < 1024; ++i)
    {
        if (!src->tables[i])
            continue;

        if (kernel_directory->tables[i] == src->tables[i])
        {
            dir->tables[i] = src->tables[i];
            dir->tablesPhysical[i] = src->tablesPhysical[i];
        }
        else
        {
            ULONG phys;
            dir->tables[i] = clone_table(src->tables[i], &phys);
            dir->tablesPhysical[i] = phys | 0x07;
        }
    }
    return dir;
}
